// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (C) 2023 Deepcomputing Technology Co., Ltd.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>
#include <pthread.h>
#include <semaphore.h>
#include <dbus/dbus.h>
#include "ec_mgr_uart_msg.h"
#include "buf.h"

// #define DEBUG 0

#define PROG_NAME "K1X-EC-COMM"
#define KEY_IDX_MAX 9999

#define UART_DEV_PATH "/dev/ttyS2"
static int uinp_fd = 0;
static int uart_fd = 0;
static sem_t buf_sem;
static sem_t dbus_sem;
static buf_t data_buf;
static int s_thread_loop = 1;
pthread_mutex_t send_mtx;
static int message_Idx [MESSAGE_TYPE_MAX];

int open_dbus(void);

static int open_key_dev()
{
    struct uinput_user_dev uinput;
    int i;
    int ret;

    uinp_fd = open("/dev/uinput", O_RDWR | O_NDELAY);
    if (uinp_fd < 0) {
        printf("[%s] Open /dev/uinput failed!\n", PROG_NAME);
        return 1;
    }

    ioctl(uinp_fd, UI_SET_EVBIT, EV_SYN);
    ioctl(uinp_fd, UI_SET_EVBIT, EV_KEY);
    ioctl(uinp_fd, UI_SET_EVBIT, EV_REL);
    ioctl(uinp_fd, UI_SET_EVBIT, EV_MSC);  

    for(i = 0; i < 256; i++) {
        ioctl(uinp_fd, UI_SET_KEYBIT, i);  
    }

    memset(&uinput,   0, sizeof(uinput));
    uinput.id.version  = 4;
    uinput.id.bustype = BUS_VIRTUAL;
    strncpy(uinput.name, "K1X UART KEY", UINPUT_MAX_NAME_SIZE);
    ret = write(uinp_fd, &uinput, sizeof(uinput));
    if (ret < 0) {
        printf("[%s] regester uinput failed!\n", PROG_NAME);
        close(uinp_fd);
        return 1;
    }

    ret = ioctl(uinp_fd, UI_DEV_CREATE);
    if (ret < 0) {
        printf("[%s] create uinput device failed!\n", PROG_NAME);
        close(uinp_fd);
        return 1;
    }

    return 0;
}

int report_key(unsigned int type, unsigned int keycode, unsigned int value)  
{  
    struct input_event key_event;  
    int ret;  
      
    memset(&key_event, 0, sizeof(struct input_event));  
      
    gettimeofday(&key_event.time, NULL);  
    key_event.type = type;  
    key_event.code = keycode;  
    key_event.value = value;  
    ret = write(uinp_fd, &key_event, sizeof(struct input_event));  
    if(ret < 0) {  
        return ret;//error process.  
    }  
      
    gettimeofday(&key_event.time, NULL);  
    key_event.type = EV_SYN;  
    key_event.code = SYN_REPORT;  
    key_event.value = 0;//event status sync  
    ret = write(uinp_fd, &key_event, sizeof(struct input_event));  
    if(ret < 0) {
        return ret;//error process.  
    }  
      
    return 0;  
}  

static int open_uart()
{
    struct termios attr;
    int ret;

    uart_fd = open(UART_DEV_PATH, O_RDWR | O_NOCTTY | O_NDELAY);
    if (uart_fd < 0) {
        printf("[%s] open uart failed!\n", PROG_NAME);
        return 1;
    }

    tcgetattr(uart_fd, &attr);
#ifdef DEBUG
    printf("[%s] tcgetattr  0x%x!\n", PROG_NAME, attr.c_cflag);
#endif

    attr.c_cflag &= ~(INLCR | IGNCR | ICRNL | ISTRIP);
    attr.c_cflag &= ~(OPOST | ONLCR | OCRNL);
    /*
        ICANON--启用标准模式 (canonical mode)。允许使用特殊字符 EOF, EOL,
                EOL2, ERASE, KILL, LNEXT, REPRINT, STATUS, 和 WERASE，以及按行的缓冲。
        ECHO--回显输入字符
        ECHOE--如果同时设置了 ICANON，字符 ERASE 擦除前一个输入字符，WERASE 擦除前一个词
        ISIG--当接受到字符 INTR, QUIT, SUSP, 或 DSUSP 时，产生相应的信号
    */
    attr.c_cflag &= ~(ISIG | ECHO | ECHOE | ECHOK | ECHONL | NOFLSH );
    attr.c_cflag &= ~CRTSCTS;// 不使用硬件流控制 
    attr.c_cflag &= ~CBAUD;
    attr.c_cflag |= B115200 | CREAD;   //115200
    attr.c_cflag &= ~CSIZE;
    attr.c_cflag |= CS8 | CLOCAL;      //8Bit data
    attr.c_cflag &= ~PARENB;           //无校验
    attr.c_cflag &= ~CSTOPB;           //1bit stopbit

#ifdef DEBUG
    printf("[%s] after set flag 2 0x%x!\n", PROG_NAME, attr.c_cflag);
#endif

    ret = tcsetattr(uart_fd, TCSANOW, &attr);
    if (ret != 0) {
        printf("[%s] tcsetattr failed %d!\n", PROG_NAME, errno);
    }

    tcgetattr(uart_fd, &attr);
#ifdef DEBUG
    printf("[%s] after set flag  0x%x!\n", PROG_NAME, attr.c_cflag);
#endif

    return 0;
}

static int write_uart(char *buf, unsigned int nLen)
{
    pthread_mutex_lock(&send_mtx);
    write(uart_fd, buf, nLen);
    //flush(uart_fd); //TODO:
    pthread_mutex_unlock(&send_mtx);

    return 0;
}

void ec_led_control(E_LED led, E_LED_MODE state)
{
    K1X_meesage_data_t msgSend;
    int keycode;
    int ret;
    char bufSend[32];

    if (led == EM_LED_BLUE)
        msgSend.type = MESSAGE_TYPE_LED_POWER;
    else
        msgSend.type = MESSAGE_TYPE_LED_CHARGE;
    msgSend.value = state;

    memset(bufSend, 0x00, sizeof(bufSend));
    ret = build_message(bufSend, sizeof(bufSend), &msgSend);
    if (ret != 0) {
        return;
    }
#ifdef DEBUG
    printf("[%s] send %s\n", PROG_NAME, bufSend);
#endif

    write_uart(bufSend, strlen(bufSend));
    return;
}

void printstr(char *str)
{
    int i = 0;

    for (i = 0; i < strlen(str); i++) {
        printf("%02x ", str[i]);
    }
    printf("end\r\n");
}

static void* MAIN_LOOP_Recv(void* param)
{
    char *addr;
    char *addrtmp;
    char bufRecv[32];
    unsigned int nPeLen;
    int ommit;
    int ret;
    int i;
    fd_set fds;
    struct timeval tv;

    while(!s_thread_loop) {
        FD_ZERO(&fds);
        FD_SET(uart_fd, &fds);
        tv.tv_sec = 0;
        tv.tv_usec = 50*1000; /*50ms*/
        if (select(uart_fd + 1, &fds, NULL, NULL, &tv) <= 0) {
#ifdef DEBUG
            printf("[%s] select fail, %d!\n", PROG_NAME, errno);
#endif
            continue;
        }

        if (!FD_ISSET(uart_fd, &fds)) {
            continue;
        }

        ommit = 0;
        nPeLen = 32;
        addrtmp = NULL;
        for (i = 0; i < 3; i++) {
            bufGetWriteBufAddr(&data_buf, &addr, &nPeLen);
            if (nPeLen > 20) {
                /*get buffer*/
                break;
            }
            if (addrtmp != NULL && addrtmp == addr) {
                /*buffer full, need ommit data*/
                ommit = 1;
                break;
            }
            addrtmp = addr;

            /*try to skip the end buffer*/
            bufSkipBuf(&data_buf, nPeLen);
        }

        if (nPeLen < 20 || ommit == 1) {
            addr = bufRecv;
            nPeLen = sizeof(bufRecv);
        }

        ret = read(uart_fd, addr, nPeLen);
        if (ret <= 0 ) {       
            continue;
        }
            //printstr(addr);

        if (ommit == 0) {
#ifdef DEBUG
            printf("[%s] read: %s\n", PROG_NAME, addr);
#endif
            bufPushData(&data_buf, ret);
            /*write to data*/
            sem_post(&buf_sem);
        } else {
            /*TODO: CLEANUP DATABUF*/
#ifdef DEBUG
            printf("[%s] ommit!\n", PROG_NAME);
#endif
            bufClean(&data_buf);
        }
    }

    return NULL;
}

static int dealwithMsg(K1X_meesage_data_t *pmsgRecv)
{
    K1X_meesage_data_t msgSend;
    int keycode;
    int ret;
    int sendkey = 0;
    char bufSend[32];

    msgSend.type = MESSAGE_TYPE_UNKNOWN;
    if (pmsgRecv->value < message_Idx[pmsgRecv->type] && (message_Idx[pmsgRecv->type] - pmsgRecv->value < KEY_IDX_MAX/2)
        && pmsgRecv->value != 0) {
        /*ommit data*/
        return 1;
    }

    message_Idx[pmsgRecv->type] = pmsgRecv->value;
    switch(pmsgRecv->type) {
        case MESSAGE_TYPE_POWEROFF:
            keycode = KEY_POWER;
            sendkey = 0;
            msgSend.type = MESSAGE_TYPE_OK;
            break;
        default:
            break;
    }

    if (msgSend.type == MESSAGE_TYPE_UNKNOWN) {
        return 1;
    }
    
    /*deal with key message*/
    if (sendkey) {
    #ifdef DEBUG
        printf("[%s] send key %d\n", PROG_NAME, keycode);
    #endif
        report_key(EV_KEY, keycode, 1);
        report_key(EV_KEY, keycode, 0);
    }

    return 0;
}

static void * MAIN_LOOP_Deal(void *param)
{
    char *addr;
    unsigned int nskip;
    unsigned int nsize;
    int i;
    K1X_meesage_data_t msgRecv;

    for (i = 0; i < MESSAGE_TYPE_MAX; i++)
        message_Idx[i] = 0;

    while(!s_thread_loop) {
        nsize = 1*1024;
        bufGetReadBufAddr(&data_buf, &addr, &nsize);
        if (nsize <= 0) {
            /*No data, wait sem*/
            sem_wait(&buf_sem);
            continue;
        }

        do {
            nskip = 0;
            msgRecv.type = MESSAGE_TYPE_UNKNOWN;
            msgRecv.value = 0;
            parse_message(addr, nsize, &msgRecv, &nskip);
            if (nskip > 0) {
                bufSkipData(&data_buf, nskip);
                nsize -= nskip;
                addr += nskip;
            }
            dealwithMsg(&msgRecv);
        } while (nsize > 0);
    }

    return NULL;
}
 
extern void* dbus_thread(void* param);
int MAIN_LOOP()
{
    pthread_t recvThd;
    pthread_t dealThd;
    pthread_t dbusThd;

    sem_init(&buf_sem, 0, 0);
    sem_init(&dbus_sem, 0, 0);
    bufInit(&data_buf, 1*1024);
    pthread_mutex_init(&send_mtx, NULL);

    pthread_create(&recvThd, NULL, MAIN_LOOP_Recv, NULL);
    pthread_create(&dealThd, NULL, MAIN_LOOP_Deal, NULL);
    pthread_create(&dbusThd, NULL, dbus_thread, NULL);

    pthread_join(recvThd, NULL);
    pthread_join(dealThd, NULL);
    pthread_join(dbusThd, NULL);

    sem_destroy(&buf_sem);
    sem_destroy(&dbus_sem);
    bufDeinit(&data_buf);
    pthread_mutex_destroy(&send_mtx);

    return 0;
}

int main(void)
{
    int ret;

    ret = open_key_dev();
    if (ret != 0) {
        printf("[%s] open_key_dev failed!\n", PROG_NAME);
    }
    ret = open_uart();
    if (ret != 0) {
        printf("[%s] open uart failed!\n", PROG_NAME);
    }
    ret = open_dbus();
    if (ret != 0) {
        printf("[%s] open dbus failed!\n", PROG_NAME);
    }

    MAIN_LOOP();

    /*Deal with POWER OFF*/
    sleep(100);  //sleep 100s

    return 0;
}
