// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (C) 2023 Deepcomputing Technology Co., Ltd.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>
#include <pthread.h>
#include <semaphore.h>
#include <dbus/dbus.h>
#include "ec_mgr_uart_msg.h"
#include <stdbool.h>
#include <syslog.h>
#include "buf.h"

// #define DEBUG 0
#define DBUS_NAME "k1.ec.comm"
#define DBUS_PATH "/k1/ec/comm"
#define DBUS_INTERFACE "ec.control.interface"

#define MSG_GETVERSION  9

static char * message_contents[] = {
    "ledpower_off",
    "ledpower_on",
    "ledpower_fblink",
    "ledpower_breath",
    "ledpower_release",
    "ledcharge_off",
    "ledcharge_on",
    "ledcharge_fblink",
    "ledcharge_breath",
    "ledcharge_release",
    ""
};

static char data[32];
static DBusConnection* connection;
static DBusHandlerResult dbus_handle_message(DBusConnection* connection, DBusMessage* message, void* user_data);

void ec_led_control(E_LED led, E_LED_MODE state);

int open_dbus(void)
{
    const char *object_path = DBUS_PATH;
    const char *interface_name = DBUS_INTERFACE;    
    DBusError error;
    int ret;

    openlog("ECLog", LOG_CONS | LOG_PID, 0);

    dbus_error_init(&error);

    connection = dbus_bus_get(DBUS_BUS_SYSTEM, &error);
    if (!connection) {
        fprintf(stderr, "Failed to open connection to bus: %s\n", error.message);
        dbus_error_free(&error);
        closelog();
        return EXIT_FAILURE;
    }

    static const DBusObjectPathVTable ec_vtable = {
        .message_function = dbus_handle_message,
    };
    ret = dbus_bus_request_name(connection, DBUS_NAME, DBUS_NAME_FLAG_REPLACE_EXISTING, &error);

    if (dbus_error_is_set(&error)) {
        fprintf(stderr, "Failed to request name: %s\n", error.message);
        dbus_error_free(&error);
        syslog(LOG_ERR|LOG_USER, "%s:%d %s\r\n", __FILE__, __LINE__, error.message);
        closelog();
        return EXIT_FAILURE;
    }
    if (DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER != ret) { 
        fprintf(stderr, "Not Primary Owner (%d)\n", ret);
        return EXIT_FAILURE;
    }
    dbus_connection_register_object_path(connection, object_path, &ec_vtable, NULL);
    dbus_connection_register_fallback(connection, object_path, &ec_vtable, NULL);

    char dbus_conn[100];
    snprintf(dbus_conn, 100, "type='signal',interface='%s'", DBUS_INTERFACE);
    dbus_bus_add_match(connection, dbus_conn, &error);

    if (dbus_error_is_set(&error)) {
        dbus_error_free(&error);
        closelog();
	return -1;
    }

    return 0;
}

static DBusHandlerResult dbus_handle_message(DBusConnection* connection, DBusMessage* message, void* user_data)
{
    int i = 0;
    int message_type = dbus_message_get_type(message);
    if (message_type == DBUS_MESSAGE_TYPE_METHOD_CALL) {

        const char *interface = dbus_message_get_interface(message);
        const char* method = dbus_message_get_member(message);
        char* p;

        if (strcmp(interface, DBUS_INTERFACE) == 0) {
            for (i = 0; i < (int)sizeof(message_contents); i++) {
                if (strcmp(method, message_contents[i]) == 0) {
                    switch (i)
                    {
                    case 0 ... 4:
                        ec_led_control(EM_LED_BLUE, i);
                        break;

                    case 5 ... 9:
                        ec_led_control(EM_LED_GREEN, i - 5);
                        break;
                    }
                    sprintf(data, "led %d", i);
                    p = data;
                    
                    DBusMessage* reply = dbus_message_new_method_return(message);
                    dbus_message_append_args(reply, DBUS_TYPE_STRING, &p, DBUS_TYPE_INVALID);
                    dbus_connection_send(connection, reply, NULL);
                    dbus_connection_flush(connection);
                    dbus_message_unref(reply);

                    break;
                }
            }
        }
    }
    return DBUS_HANDLER_RESULT_HANDLED;
}

void * dbus_thread(void *param)
{
    DBusMessage* msg;
    // loop, testing for new messages
    while (true) {
        dbus_connection_read_write_dispatch(connection, -1);
    }
}

